"""Library containing the complementary mathematical functions necessary for the operation of the code.

	Autor: Javier Díez Sierra
	Fechas 06-02-2017
    
    Updated by Salvador Navas Fernández
    Date 01-07-2021
"""



import sys
import matplotlib.pyplot as plt

import pandas as pd
import numpy as np
import datetime
from datetime import timedelta

def allmonths(dataframe):
    """Paso el dataframe a 12 columnas con 12 meses"""
    
    dataframe_meses=pd.DataFrame(index=dataframe.index, columns=np.arange(1,13))
    if len(dataframe.columns)==12:
        dataframe_meses.loc[:,:] = dataframe.values
    else:
        for i,ii in enumerate(dataframe.columns):
            for j in list(map(int, ii[1:-1].split(','))):
                dataframe_meses.loc[:,j] = dataframe.loc[:,ii]

   
    return dataframe_meses


def datetime2matlabdn(dt):
    mdn = dt + timedelta(days = 366)
    frac_seconds = (dt-datetime.datetime(dt.year,dt.month,dt.day,0,0,0)).seconds\
    /(24.0 * 60.0 * 60.0)
    frac_microseconds = dt.microsecond / (24.0 * 60.0 * 60.0 * 1000000.0)
    return mdn.toordinal() + frac_seconds + frac_microseconds



def datetime2matlabdnJavi(pdt, dt):
    '''
    Funcion que passa de un vectors de pandas a una lista de tiempo en formato ordinal
    pdt=pd.date_range
    dt = resolución days/hours/minutes/seconds
    '''
    
    if dt=='days':
        dt_ordinal=1
    elif dt=='hours':
        dt_ordinal=1/24
    elif dt=='minutes':
        dt_ordinal=1/(24*60)
    elif dt=='seconds':
        dt_ordinal=1/(24*60*60)
        
    d1 = datetime.datetime(pdt.year[0], pdt.month[0], pdt.day[0])# start date
    tt_ordinal_aux=list()
    
    dtt=0
    for i in range(len(pdt)):
        if i ==0:
            #tt_ordinal_origen=(d1).toordinal()
            tt_ordinal_origen=datetime2matlabdn(d1)
            tt_ordinal_aux.append(tt_ordinal_origen)
        else:
            tt_ordinal_aux.append(tt_ordinal_origen+dtt)
            
        dtt=dtt+dt_ordinal
        
    return tt_ordinal_aux

def func(x, a, b, c):
    """Funcion a la que voy a ajustar los valores de correlación espacial"""

    return a * np.exp(-b * x) + c

def distancia_f(x1, y1, x2, y2):
    dist=((x1-x2)**2 + (y1-y2)**2)**0.5
    return dist
def haversine(lon1, lat1, lon2, lat2):
    from math import radians, cos, sin, asin, sqrt
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a)) 
    km = 6367 * c
    return km

    model.params
    model.plot_summary()
    RETURN_PERIODS.return_periods=model.return_periods
    RETURN_PERIODS.return_values=model.return_values
    plt.show()

def IDW_f(xx, yy, zz, x, y, betha):
    ##xx, yy --> Coordenadas de las estaciones
    ##zz --> variable que quieres interpolar
    ##x, y --> punto donde quieres obtener el resultado
    ## betha--> exponente de peso. Cuanto mayor, más peso se le asigna a los puntos cercanos
    dist=list()
    for i in range(len(xx)):
        dist.append(haversine(xx[i], yy[i], x, y))
    dist=np.array(dist)
    weights=(dist**(-betha))/np.sum(dist**(-betha))
    
    return np.sum(zz*weights)

def compare_statistics(CAL, SIM):
    
    stats_obs=allmonths(CAL.statististics_Real)
    stats_fit=allmonths(CAL.statististics_Fit)
    stats_sim=allmonths(SIM.statististics_Simulated)
    
    RK=int(np.ceil(np.sqrt(len(stats_obs.index)))); 
    if len(stats_obs.index) % RK == 0:
        CK = int(len(stats_obs.index)/RK)
    else:
        CK = int(np.ceil(np.sqrt(len(stats_obs.index))))
    fig, axes = plt.subplots(RK, CK, figsize=(15, 15))
    for i, ii in enumerate(stats_obs.index):
        Data_sta=pd.DataFrame(index=np.arange(1, 13))
        Obs=list(); Fit=list(); Sim=list();
        for m in np.arange(1, 13):
            Obs.append(stats_obs[m].loc[ii])
            Fit.append(stats_fit[m][ii])
            Sim.append(stats_sim[m].loc[ii])

        Data_sta['Obs']=Obs
        Data_sta['Fit']=Fit
        Data_sta['Sim']=Sim
        #Ploteo
        row = i // CK
        col = i % CK
        ax=axes[row,col]
        ax.set_title(ii)
        ax.grid(True)
        #Data_sta.plot(ax=axes[row,col])
        legnd=['Obs', 'Fit', 'Sim']
        pp=Data_sta['Obs'].plot(style='k--', lw=2,  ax=axes[row,col])
        Data_sta['Fit'].plot(style='bs', ax=axes[row,col])
        Data_sta['Sim'].plot(style='r^', ax=axes[row,col])
        pp.legend(legnd, loc='best', fontsize=15)
        if ii[0:2]==('fi' or 'au'):
            ax.set_ylim(0, 1)

def exceedence_probability(Serie_Observed, Serie_Simulated, temporal_resolution):
    # Observed
    aux=Serie_Observed.values; aux=aux[aux>=0]
    s_obs = np.sort(aux, axis=None)[::-1]
    exc_obs = np.arange(1.,len(s_obs)+1) / len(s_obs)

    # Simulated
    if temporal_resolution == 'd':
        units = 'mm/day'
    elif temporal_resolution == 'h':
        units = 'mm/hour'

    aux = Serie_Simulated.values; aux=aux[aux>=0]
    s_sim = np.sort(aux, axis=None)[::-1]
    exc_sim = np.arange(1.,len(s_sim)+1) / len(s_sim)
    
    fig, ax = plt.subplots(figsize=(10, 5))

    ax.plot(exc_obs*100, s_obs, '.k', markersize=8,  label='Obs')
    ax.plot(exc_sim*100, s_sim, 'r',  label='Sim')
    ax.set_xlabel("Exceedence Probability", fontsize=13)
    ax.set_ylabel(units, fontsize=13)
    ax.set_xscale('log')
    ax.grid(True)
    ax.tick_params(labelsize=13)
    ax.legend(fontsize = 15)
 

	




'''
Library containing classes for simulating time series from the model.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''



from NSRP.MathematicalPropertiesNSRP import *
from NSRP.utils import *
from NSRP.libs_NSRP import *
from NSRP.inputs_simulation import *
from NSRP.outputs_simulation import *
import time

class Simulation(object):
    def __init__(self,hiperparams):
        self.hiperparams = hiperparams
        
    def __call__(self, params_cal):
   
        statististics_simulacion_dataframe=pd.DataFrame(index=self.hiperparams.statistics_name,columns=self.hiperparams.Seasonality_str)
        
        Dataframe_parametros = inputs_simulation(params_cal).params_cal
        
        
        Dataframe_parametros=allmonths(Dataframe_parametros)
        print('')
        print('')
        print('#'*80)
        print('Synthetic simulation')
        print('')
        print('')
		## We start with synthetic simulation and statistical calculations.
        
        if self.hiperparams.process=='normal':
            [Dataframe_simulacion_unido_hour_, Dataframe_simulacion_unido_day_, Intensidad_cellss, Duracion_horas_cellss]=\
            NSRP_simulation1(Dataframe_parametros, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, self.hiperparams.process, self.hiperparams.Seasonality)
            print('Total cumulative rainfall - Analytical estimation = ' + "{:15.2f}".format(np.sum(Intensidad_cellss*Duracion_horas_cellss)))
            print('Total cumulative rainfall -             Simulated = ' + "{:15.2f}".format(np.sum(Dataframe_simulacion_unido_day_.values)))
            
        elif self.hiperparams.process=='storms':

            Dataframe_parametros1=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'xi', 'betha'],columns=Dataframe_parametros.columns)
            for i in Dataframe_parametros.columns:
                Dataframe_parametros1.loc[:,i]=Dataframe_parametros.loc[:,i].values[0:5]

            Dataframe_parametros2=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'xi', 'betha'],columns=Dataframe_parametros.columns)
            for i in Dataframe_parametros.columns:
                Dataframe_parametros2.loc[:,i]=Dataframe_parametros.loc[:,i].values[5:]

            [Dataframe_simulacion_unido_hour_1, Dataframe_simulacion_unido_day_1, Intensidad_cellss1, Duracion_horas_cellss1]=\
            NSRP_simulation1(Dataframe_parametros1, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, 'normal', self.hiperparams.Seasonality)

            [Dataframe_simulacion_unido_hour_2, Dataframe_simulacion_unido_day_2, Intensidad_cellss2, Duracion_horas_cellss2]=\
            NSRP_simulation1(Dataframe_parametros2, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, 'normal', self.hiperparams.Seasonality)

			##Combino los dos processos
            Dataframe_simulacion_unido_hour_=pd.DataFrame(index=Dataframe_simulacion_unido_hour_1.index)
            Dataframe_simulacion_unido_hour_['Rain']=Dataframe_simulacion_unido_hour_1.values+Dataframe_simulacion_unido_hour_2.values
            Dataframe_simulacion_unido_day_=pd.DataFrame(index=Dataframe_simulacion_unido_day_2.index)
            Dataframe_simulacion_unido_day_['Rain']=Dataframe_simulacion_unido_day_1.values+Dataframe_simulacion_unido_day_2.values

            print('Total cumulative rainfall - Analytical estimation - Storm 1 = ' + "{:15.2f}".format(np.sum(Intensidad_cellss1*Duracion_horas_cellss1))) 
            print('Total cumulative rainfall - Analytical estimation - Storm 2 = ' + "{:15.2f}".format(np.sum(Intensidad_cellss2*Duracion_horas_cellss2)))  

            print('Total cumulative rainfall - Analytical estimation = ' + \
            "{:15.2f}".format(np.sum(Intensidad_cellss1*Duracion_horas_cellss1)+np.sum(Intensidad_cellss2*Duracion_horas_cellss2)))
            print('Total cumulative rainfall -             Simulated = ' + "{:15.2f}".format(np.sum(Dataframe_simulacion_unido_day_.values)))    


        #statististics_dataframe=allmonths(statististics_dataframe)
        #print('')
        #print('')
        #print('#'*80)
        #print('Results Validation')
        #print('')
        #print('')
        
        ## Real, adjusted and simulated statistical calculations.

        for pri, prii in enumerate(self.hiperparams.Seasonality):
            if self.hiperparams.temporal_resolution == 'd':
                Datos=Dataframe_simulacion_unido_day_.copy()
            elif self.hiperparams.temporal_resolution == 'h':
                Datos=Dataframe_simulacion_unido_hour_.copy()
            Datos[Datos['Rain']<0]=np.nan
            Datos[Datos['Rain']<0]=np.nan

            if len(self.hiperparams.Seasonality)==12:
             
                Datos['Estacionalidad']=Datos['Rain']*np.nan
                pos=np.where(Datos.index.month == prii); pos=pos[0]
                Datos['Estacionalidad'].iloc[pos]=Datos['Rain'].iloc[pos]
                Pluvio_GS = Datos['Estacionalidad'][Datos['Estacionalidad']>=0]
                Pluvio_GS[Pluvio_GS<0.001]=0
                Datos=Pluvio_GS.astype(float)
                
            else:
  
                Datos['Estacionalidad']=Datos['Rain']*np.nan
                for i, ii in enumerate(prii):
                    pos=np.where(Datos.index.month == ii); pos=pos[0]
                    Datos['Estacionalidad'].iloc[pos]=Datos['Rain'].iloc[pos]
                Pluvio_GS = Datos['Estacionalidad'][Datos['Estacionalidad']>=0]
                Pluvio_GS[Pluvio_GS<0.001]=0
                Datos=Pluvio_GS.astype(float)



            statististics_values_sintetico=calculate_statistics(Datos,self.hiperparams.statistics_name,self.hiperparams.temporal_resolution)
            statististics_simulacion_dataframe.loc[:,str(prii)]=statististics_values_sintetico  

        #statististics_simulacion_dataframe=allmonths(statististics_simulacion_dataframe)
        
        results = outputs_simulation(Dataframe_simulacion_unido_day_,Dataframe_simulacion_unido_hour_,statististics_simulacion_dataframe, self.hiperparams.temporal_resolution)

        return results
        
    # def save_files(self,results,path_output_files):
        
    #     results.Hourly_Simulation.to_csv(path_output_files+'Time_serie_hourly_simulated.csv')
    #     results.statististics_Simulated.to_csv(path_output_files+'statistic_hourly_simulated.csv')
    #     results.Daily_Simulation.to_csv(path_output_files+'Time_serie_day_simulated.csv')



'''
Library containing classes for reading yaml configuration files for calibration and simulation.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''



import yaml
import ast

class Calibration(object):
    """This function allows to obtain the hyperparameters to start the calibration of the point model and to check if all parameters are correct.
    
    Parameters
    ----------
    file_yml : String. 
        Directory where the Yml file with the configuration to perform the model calibration is located.
      
    """
    def __init__(self,file_yml):
        with open(file_yml) as file:
            hiper_params = yaml.load(file, Loader=yaml.FullLoader)
        self.Seasonality_type        = hiper_params['Seasonality_type']
        self.Seasonality_user        = hiper_params['Seasonality_user']
        self.temporal_resolution     = hiper_params['temporal_resolution']
        self.process                 = hiper_params['process']
        self.statistics_name         = hiper_params['statistics_name']
        self.weights                 = hiper_params['weights']
        self.number_iterations       = hiper_params['number_iterations']
        self.number_bees             = hiper_params['number_bees']
        self.number_initializations  = hiper_params['number_initializations']
        self.time_between_storms     = hiper_params['time_between_storms']
        self.number_storm_cells      = hiper_params['number_storm_cells']
        self.cell_duration           = hiper_params['cell_duration']
        self.cell_intensity          = hiper_params['cell_intensity']
        self.storm_cell_displacement = hiper_params['storm_cell_displacement']
        
        
        if self.Seasonality_type == 'annual':
            Seasonality=list()
            Seasonality.append((1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ,11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type == 'seasonal':
            Seasonality=list()
            Seasonality.append((1, 2, 3))
            Seasonality.append((4, 5, 6))
            Seasonality.append((7, 8, 9))
            Seasonality.append((10, 11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type == 'monthly':
            Seasonality=list()
            Seasonality.append((1))
            Seasonality.append((2))
            Seasonality.append((3))
            Seasonality.append((4))
            Seasonality.append((5))
            Seasonality.append((6))
            Seasonality.append((7))
            Seasonality.append((8))
            Seasonality.append((9))
            Seasonality.append((10))
            Seasonality.append((11))
            Seasonality.append((12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
                
        elif self.Seasonality_type == 'user_defined':
        
            Seasonality_str = hiper_params['Seasonality_user']
            
            Seasonality = list()
            for i in self.Seasonality_user:
                Seasonality.append(ast.literal_eval(i))
                
            if type(hiper_params['Seasonality_user'][0])!=str:
                raise Exception ('The definition of the months that make up the user-defined seasonality is incorrect, check that it is a string.')
                
        else:
            raise Exception ('The name of the seasonality type is not correct, check that the name is among those listed below: annual, monthly, seasonal, user_defined')       
        self.Seasonality_str = Seasonality_str        
        self.Seasonality = Seasonality
        
        
            
        if self.temporal_resolution != 'h' and self.temporal_resolution != 'd':
            raise Exception ("The time resolution of the point model is poorly defined. It should be 'h' or 'd'.")
        
        if self.process != 'storms' and self.process != 'normal':
            raise Exception ("The process type is wrongly defined, it should be 'storms' or 'normal'.")
            
            
        statistic_list = ['mean_h', 'var_h', 'autocorr_l_h','fih_h', 'fiWW_h', 'fiDD_h', 'M3_h']
        
        for i in self.statistics_name:
            if i.split('_')[0]== 'autocorr':
                st =  i.split('_')[0]+'_'+'l'+'_'+'h'
                if st not in statistic_list:
                    raise Exception ('The '+i+' statistic has not been correctly defined.')
            else:
                st = i.split('_')[0]+'_'+'h'
                if st not in statistic_list:
                    raise Exception ('The '+i+' statistic has not been correctly defined.')
        
            
            
            
        
        
        
class Simulation(object): 
    """This function allows to obtain the hyperparameters to start the simulation of the point model and to check if all parameters are correct.
        
        Parameters
        ----------
        file_yml : String. 
            Directory where the Yml file with the configuration to perform the model simulation is located.
    """
    def __init__(self,file_yml_sim): 
        with open(file_yml_sim) as file:
            hiper_params = yaml.load(file, Loader=yaml.FullLoader)
        self.Seasonality_type        = hiper_params['Seasonality_type']
        self.Seasonality_user        = hiper_params['Seasonality_user']
        self.statistics_name         = hiper_params['statistics_name']
        self.process                 = hiper_params['process']
        self.temporal_resolution     = hiper_params['temporal_resolution']
        self.year_ini                = hiper_params['year_ini']
        self.year_fin                = hiper_params['year_fin']
        
        if self.Seasonality_type=='annual':
            Seasonality=list()
            Seasonality.append((1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ,11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type=='seasonal':
            Seasonality=list()
            Seasonality.append((1, 2, 3))
            Seasonality.append((4, 5, 6))
            Seasonality.append((7, 8, 9))
            Seasonality.append((10, 11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type =='monthly':
            Seasonality=list()
            Seasonality.append((1))
            Seasonality.append((2))
            Seasonality.append((3))
            Seasonality.append((4))
            Seasonality.append((5))
            Seasonality.append((6))
            Seasonality.append((7))
            Seasonality.append((8))
            Seasonality.append((9))
            Seasonality.append((10))
            Seasonality.append((11))
            Seasonality.append((12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
                
        elif self.Seasonality_type =='user_defined':
        
            Seasonality_str = hiper_params['Seasonality_user']
            
            Seasonality = list()
            for i in self.Seasonality_user:
                Seasonality.append(ast.literal_eval(i))
                
            if type(hiper_params['Seasonality_user'][0])!=str:
                raise Exception ('The definition of the months that make up the user-defined seasonality is incorrect, check that it is a string.')
                
        else:
            raise Exception ('The name of the seasonality type is not correct, check that the name is among those listed below: annual, monthly, seasonal, user_defined')
                
        self.Seasonality_str = Seasonality_str        
        self.Seasonality = Seasonality


'''
Library containing the main functions for the execution of the model.

	Author: Javier Díez Sierra
	Date 06-02-2017
    
    Updated by Salvador Navas Fernández
    Date 05-07-2021
    
'''

import pandas as pd

class inputs_simulation (object):
    def __init__(self,inputs_cal):
            
        if type(inputs_cal) != str:
            self.params_cal = inputs_cal.Fitted_parameters
        elif type(inputs_cal) == str:
            self.params_cal = pd.read_csv(inputs_cal,index_col=0)
        else:
            raise Exception


#!/usr/bin/env python

from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='STNSRPM',
      version='0.0.1',
      description='🌎 Scripts and information to synthetic generation of precipitation based on Point Processes.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Javier Diez Sierra',
      author_email='javier.diez@unican.es',
      maintainer       = 'Manuel del Jesus Peñil',
      maintainer_email = 'manuel.deljesus@unican.es',
      url = 'https://github.com/navass11/STNSRPM/tree/Crear-ejecutable',
      packages = ['NSRP'],
      include_package_data = True,
      python_requires='>=3.7, <4',
      install_requires=[
          'numpy',
          'pandas',
          'scipy',
          'datetime',
          'matplotlib',
          'pyyaml',
      ],
      extras_require={'plotting': ['matplotlib>=2.2.0', 'jupyter','jupyterlab']}
     )